import * as DOM from './dom.js'
import DATA_INTERFACE from "./dataInterface";
import DATA_STORE from "./dataStore";

export default class ROUTE {

    constructor(oProps) {

        this.sTitle = null;

        this.cAPP = oProps.cAPP;

        this.sRoutePath = oProps.sRoutePath;

        this.oParentRoute = (oProps.oParentRoute) ? oProps.oParentRoute : null;
        this.oNavState = (oProps.oNavState) ? oProps.oNavState : null;

        this.dRouteContainer = (this.oParentRoute && this.oParentRoute.childViewLocation) ? this.oParentRoute.childViewLocation : this.cAPP.dAppBody; 

        console.log("Route Container!");
        console.log(this.oParentRoute);
        console.log(this.dRouteContainer);

        // Set active route flag on container element
        this.dRouteContainer.dataset.ar = this.sRoutePath;

        // Selector to find the child view container (during post render)
        this.sChildViewContainerSelector = null;

        // Flag to skip adding an additional template framgement wrapper
        this.bSkipTemplateFragementWrapper = false;

        this.dChildViewContainer = null;
      
        this.cDataInterface = new DATA_INTERFACE();
        this.cDataInterface.init();

        this.oLifeCycleNames = {
            init: 'init',
            reload:'reload',
            preRender:'preRender',
            render:'render',
            postRender:'postRender',
            unload:'unload',            
        }

        this._sGuardCycle = this.oLifeCycleNames.init;

        this.cDataStore = new DATA_STORE(); 

        this.sRouteStyles = null;
        this.sRouteTemplate = null;

        // Properties used by applications template renderer
        this.cRenderer = null;
        this.aoRendererTemplate = null;
        this.oRendererTemplateList = null;
        this.sRendererInitialTemplateName = null;
        this.oRendererData = null;

        this.dRouteTemplateFragment = null;

    }

    get childViewLocation(){

        if (this.dChildViewContainer) {

            return this.dChildViewContainer;
        }
        else {

            console.log("No child view container was defined!");
        }
    }

    set sGuardCycle(sCycle){
        if(sCycle && this.oLifeCycleNames.hasOwnProperty(sCycle)){
            this._sGuardCycle = sCycle;
        }
    }

    _guard(){
        if(this.guard){
            let guardResponse = this.guard();   
            
            if(guardResponse === true){                
                /*DEBUG: START*/
                console.log('Guard tripped - ABORT');
                /*DEBUG: END*/
            }
            else if(typeof guardResponse == "string"){    // if string - nav there
                cAPP.navTo(guardResponse);                
            }
        }     
    }

    _checkGuard(sCycle){
        if(sCycle == this._sGuardCycle){
            this._guard();
        }
    }

    //================
    //Lifecycle methods
    //================

    _init () {
        DOM.clearElements(this.dRouteContainer);
        
        this._setRouteStyles();

        this._callChildMethod(this.oLifeCycleNames.init).then(()=>{
            this._checkGuard(this.oLifeCycleNames.init);
            this._preRender();
        })  
    
    }

    _preRender(){        
        this._createRouteTemplateFragment(); 

        if(this.dRouteTemplateFragment){
                // Call updateLinks here as the content should be present in the screenFragment. 
                this._updateLinks();
                this._bindActions();

                this._bindChildViewContainer()
            }

        // Call next lifecycle method            
        // PreRender does not allow for child methods to be called
        this._checkGuard(this.oLifeCycleNames.preRender);
        this._render();
    }

    _render () {  
        this._renderRouteTemplateFragment()

        // Call next lifecycle method            
        // Render does not allow for child methods to be called. 
        this._checkGuard(this.oLifeCycleNames.render);
        this._postRender();
        
    }

    _postRender() {

        if (this.sChildViewContainerSelector) {

            let adViewContainer = this.dRouteContainer.querySelectorAll(this.sChildViewContainerSelector);

            if (adViewContainer.length === 1) {

                this.dChildViewContainer = adViewContainer[0];
            }
            else {

                console.error(`Child view container selector is not unique: ${this.sChildViewContainerSelector}`);
            }

        }
        else {

            this.dChildViewContainer = this.dRouteContainer;
        }

        // Call child method and next lifecycle method
        this._callChildMethod(this.oLifeCycleNames.postRender).then(()=>{
            this._checkGuard(this.oLifeCycleNames.postRender);            
        })  
    }

    _unload(){
        /*DEBUG: START*/
        console.log('ROUTE - Unload - '+ this.sTitle);
        /*DEBUG: END*/
        this._callChildMethod('unload');
    }


    //================
    // Other utility methods
    //================

    _setRouteStyles(){
        if(this.sRouteStyles){
            this.dStyleContainer = document.createElement('style');
            this.dStyleContainer.dataset.rc = this.sRoutePath + "-style";
            this.dStyleContainer.innerHTML = this.sRouteStyles;
            this.dRouteContainer.prepend(this.dStyleContainer);
        }
    }

    async _callChildMethod(sMethod){
        if(this[sMethod] && typeof this[sMethod] == "function"){
            this[sMethod]();
        }
    }

    _renderRouteTemplateFragment() {

        if(this.dRouteContainer){

            if(this.dRouteTemplateFragment){
            
                if (this.dStyleContainer){
                    this.dRouteContainer.appendChild(this.dStyleContainer);
                }

                this.dRouteTemplateFragment.dataset.rc = this.sRoutePath + "-tf";

                this.dChildViewContainer.dataset.rc = this.sRoutePath + "-cvc";
            
                this.dRouteContainer.appendChild(this.dRouteTemplateFragment);
            }  
            else if(this.dStyleContainer){

                let dChildViewContainer = document.createElement('div');
                dChildViewContainer.dataset.rc = this.sRoutePath + "-cvc";

                this.dChildViewContainer = dChildViewContainer;

                this.dRouteContainer.appendChild(this.dStyleContainer);
                this.dRouteContainer.appendChild(dChildViewContainer);
            }     
        }
    }

    _createRouteTemplateFragment(){
        if (this.sRouteTemplate) {

            // Create dom fragment of template. 
            this.dRouteTemplateFragment = document.createElement('div');    //Using a div to perserve reference and not need to recreate every time.     
            this.dRouteTemplateFragment.innerHTML = this.sRouteTemplate;

            if (this.bSkipTemplateFragementWrapper) {
                this.dRouteTemplateFragment = this.dRouteTemplateFragment.firstElementChild;
            }
        }   
        else if (this.oRendererTemplateList || this.aoRendererTemplate) {
            this.dRouteTemplateFragment = document.createElement('div');

            if (this.oRendererTemplateList) {

                if (this.sRendererInitialTemplateName) {

                    // Create a renderer instance
                    this.cRenderer = this.cAPP.newRenderer(this.oRendererTemplateList);
    
                    this.dRouteTemplateFragment.appendChild(this.cRenderer.generate(this.sRendererInitialTemplateName, this.oRendererData));
                }
                else {
                    console.error("When providing multiple local templates, you must delcare the name of the template to use as the screen template. Do this using this.sRendererInitialTemplateName in the screen constructor.");
                }
            }
            else {

                // Create a renderer instance
                this.cRenderer = this.cAPP.newRenderer(this.aoRendererTemplate);

                this.dRouteTemplateFragment.appendChild(this.cRenderer.generate("_screen", this.oRendererData));
            }

        } 
        else {

            //console.error(`A static tpl.html, .plates, or PLATES_TEMPLATES should be provided`);
        }   

        
    }

    _bindChildViewContainer(){
        if(!this.dChildViewContainer){
            if(this.sChildViewContainerSelector){
                let dChildViewContainer = this.dRouteTemplateFragment.querySelector(this.sChildViewContainerSelector);

                if(dChildViewContainer){
                    this.dChildViewContainer = dChildViewContainer;
                }
                else{
                    console.error("Screen selector did not resolve to an element within the route's rendered content.");
                }
            }
            else{
                console.error("When generating route content, a sChildViewContainerSelector must be provided or dChildViewContainer must be manually defined.");
            }
        }
    }

    _updateLinks(){
        let adLinks = this.dRouteTemplateFragment.querySelectorAll('a');

        for (let a = 0, aLen = adLinks.length; a < aLen; a++) {
            let dLink = adLinks[a];
            
            dLink.addEventListener('click', (evt) => {
                evt.preventDefault();
                let sNavTo = evt.target.getAttribute('href');
                cAPP.navTo(sNavTo, null, evt,);
            });
        }
    }

    _bindActions(){
        this.dActionContainer = this.dRouteTemplateFragment;
        //If building html in render function there wont be a screen template fragment.
        if(!this.dRouteTemplateFragment){
            this.dActionContainer = this.dRouteContainer;
        }

        let aHasAction = this.dActionContainer.querySelectorAll("[data-action]");
        
        if(aHasAction && aHasAction.length > 0){
            for(let i =0; i < aHasAction.length; i++){
                let dHasAction = aHasAction[i];

                let sAction = dHasAction.dataset.action;
                let sActionType = dHasAction.dataset.actionType;

                if(sAction && sAction != "" && this[sAction] && sActionType && sActionType != "" && !dHasAction.onclick){
                    switch(sActionType){
                        case 'onclick':
                            //If defining a click event, should set type to button incase defaulted to submit. 
                            if(dHasAction.nodeName == "BUTTON"){
                                dHasAction.type = "button";    
                            }                            
                            dHasAction.onclick = this[sAction].bind(this);
                        break;                        
                    }
                }
            }
        }
    }  

    _registerService(sServiceName, oServiceProps){
        if(sServiceName && oServiceProps){

            if(!this.cDataInterface){
                this.cDataInterface = new DATA_INTERFACE();
                this.cDataInterface.init();
            }

            return this.cDataInterface.registerService(sServiceName, oServiceProps);
        }
        return null;
    }
 
    _registerDataStore(sDataStoreName, oDataStoreProps){
        if(sDataStoreName && oDataStoreProps){
            
            if(!this.cDataInterface){
                this.cDataInterface = new DATA_INTERFACE();
                this.cDataInterface.init();
            }

            return this.cDataInterface.registerDataStore(sDataStoreName, oDataStoreProps);
        }
        return null;
    }
}